/*
 * BallRenderer.java
 *
 * Created on February 7, 2003, 4:44 PM
 */

package ballworld;

import cs101.lang.Animate;
import cs101.lang.AnimatorThread;

import java.awt.Point;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Ellipse2D.Double;


/** Holds and animates one ball which can be rendered upon requested. Also
 * provides methods for transforming between world corrdinates (cartesian
 * with 0,0 at center) and GUI coordinates (0,0 at upper left with inverted
 * vertical scale as is the norm in Java GUI components)
 *
 * @author Patrick G. Heck, gus.heck@olin.edu
 */
public class BallRenderer {
    
    private Ball myBall;
    private BallWorld myWorld;
    private AnimatorThread at;
    private boolean selected;
    
    /** Creates a new instance of BallRenderer
     * @param b The {@link Ball} we are responsible for rendering
     * @param w The {@link BallWorld} that
     */
    public BallRenderer(Ball b, BallWorld w) {
        this.myBall = b;
        this.myWorld = w;
        if (b instanceof Animate) {
            at = new AnimatorThread((Animate)b,AnimatorThread.START_IMMEDIATELY,0,50);
        }        
    }

    /**
     * Draw the ball on the given graphics context.
     *
     * @param G  A graphics context for drawing
     */
    public void drawBall(Graphics2D G) {
        double radius = myBall.getRadius();
        double diam = radius*2;
        Point worldPoint = new Point();
        worldPoint.setLocation(myBall.getX(), myBall.getY());
        Point guiPoint = BallRenderer.transformWorld2GUI(worldPoint, myWorld);
        double x = guiPoint.getX();
        double y = guiPoint.getY();
        Ellipse2D.Double ball = 
            new Ellipse2D.Double(x-radius,y-radius,diam,diam);
        if(selected) {
            G.setPaint(Color.red);
        } else {
            G.setPaint(Color.black);
        }
        G.fill(ball);
    }
    
    /**
     * Get the ball held by this renderer.
     *
     * @return the ball contained by this renderer.
     */
    public Ball getBall() {
        return myBall;
    }
    
    /**
     * Set the selection status for this ball;
     * 
     * @param state the state to set (selected or not)
     */
    public void setSelected(boolean state) {
        selected = state;
    }
    
    /** Check whether or not the given point resides within the radius of the
     * ball.
     *
     * @param p The point to check
     * @return <code>true</code> if our ball contains this point, <code>false</code> otherwise.
     */
    public boolean containsPoint(Point p) {
        if (distanceTo(p) <= myBall.getRadius()) {
            return true;
        }
        return false;
    }
    
     /** Find the distance from our ball to the given point.
      *
      * @param p The point to measure the distance to
      * @return The distance between the given coordinate and our ball
      */
    public double distanceTo(Point p) {
        double x0 = myBall.getX();
        double y0 = myBall.getY();
        double x1 = p.getX();
        double y1 = p.getY();
        
        return Math.pow(Math.pow(x0-x1,2) + Math.pow(y0-y1,2),0.5);
        
    }
    
    /**
     * Stop the Animator thread. 
     */
    public void stop() {
        at.stopExecution();
    }
    
    /** Convert a point from world coordinates to GUI coordinates.
     * World coordinates are 0,0 at the center of the window with Y
     * increasing in the up direction (standard Cartesian coordinates).
     * GUI Coordinates are 0,0 at the upper left with Y increasing downwards
     *
     * @return A point in GUI coordinates
     * @param w The world we are converting from
     * @param p A point in world coordinates
     */
    public static Point transformWorld2GUI(Point p, BallWorld w) {
        Point newP = new Point();
        double worldH = w.getGUIMaxY();
        double worldW = w.getGUIMaxX();
        newP.setLocation(p.getX() + worldW/2,
                        worldH/2 - p.getY());
        
        return newP;
    }
    
    /** Convert a point from GUI coordinates to world coordinates.
     * World coordinates 0,0 at the center of the window with Y
     * increasing in the up direction (standard Cartesian coordinates).
     * GUI Coordinates are 0,0 at the upper left with Y increasing downwards
     *
     * @return A point in world coordinates
     * @param w The world we are converting to
     * @param p A point in GUI coordinates
     */
    public static Point transformGUI2World(Point p, BallWorld w) {
        Point newP = new Point();
        double worldH = w.getGUIMaxY();
        double worldW = w.getGUIMaxX();
        newP.setLocation(p.getX() - worldW/2,
                        worldH/2 - p.getY());
        
        return newP;
    }
}
