/*
 * BallCanvas.java
 *
 * Created on February 7, 2003, 2:24 PM
 */

package ballworld;

import java.net.URL;

import java.awt.Point;
import java.awt.Image;
import java.awt.Color;
import java.awt.Toolkit;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.Graphics2D;
import java.awt.TexturePaint;
import java.awt.MediaTracker;

import java.awt.event.KeyEvent;
import java.awt.event.KeyAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseAdapter;

import java.awt.image.BufferedImage;

import java.util.Vector;
import java.util.Iterator;



/**
 * Provide a canvas with a grassy backround to draw balls on.
 *
 * @author  Patrick G. Heck, gus.heck@olin.edu
 */
public class BallWorld extends javax.swing.JComponent implements World, Runnable {
    
    private BallRenderer selectedBall;
    private BufferedImage background;
    private Vector ballList;
    private boolean running;
    private ClassLoader cl = Thread.currentThread().getContextClassLoader();
    
    /** Creates a new instance of BallWorld */
    public BallWorld() {
        ballList = new Vector();
        URL whereImage = cl.getResource("IMAGES/grass.jpg");
        background = loadBufferedImage(whereImage);
        this.addMouseListener(new MouseAdapter() {
            public void mousePressed(MouseEvent evt) {
                processClick(evt);
            }
            public void mouseReleased(MouseEvent evt) {
                processClick(evt);
            }
        });
        this.addKeyListener( new KeyAdapter() {
            public void keyTyped(KeyEvent evt) {
                processKey(evt);
            }
        });
    }
    
    /**
     * Grab an image from a URL and stick it in a <code>BufferedImage</code>. 
     * This method is desgned to be able to handle the loading of an image 
     * from both slow and fast locations. * A <code>MediaTracker</code> is 
     * used ot ensure that no operations are perfored on the image before 
     * loading is complete.
     *
     * @param imageLoc  Where to get the image from
     * @return          The loaded image
     */
    
    private BufferedImage loadBufferedImage(URL imageLoc) {
        
        // create a place to hold the background pixmap for the board
        Image img;
        BufferedImage B;
        int H, W;
        Graphics2D G;
        img = Toolkit.getDefaultToolkit().getImage(imageLoc);
        MediaTracker wholeImg = new MediaTracker(this);
        wholeImg.addImage(img,0);
        try
        {
            wholeImg.waitForAll();
        }
        catch (InterruptedException e)
        {
            System.out.println("Image load interrupted " + 
            "for background!\n" + e);
        }
        
        H = img.getHeight(this);      // now that we have it how big is it?
        W = img.getWidth(this);
        
        // Now we have to move it into a BufferedImage...
        
        B = new BufferedImage(W,H,BufferedImage.TYPE_INT_ARGB);
        
        G = B.createGraphics();
        G.drawImage(img,0,0,this);
        
        return B;
    }
    
    /**
     * Draws the grass and all the balls.
     *
     * @param g  a graphics context to draw in
     */
    public void paint(Graphics g) {
        BufferedImage newImg = new BufferedImage(this.getWidth(),
                                                 this.getHeight(),
                                                 BufferedImage.TYPE_INT_ARGB);
        Graphics2D G = newImg.createGraphics();
        TexturePaint grass = new TexturePaint(background, 
            new Rectangle(0,0,background.getWidth(),background.getHeight()));
        Rectangle wholeCanvas = new Rectangle(0,0,this.getWidth(),this.getHeight());
        
        G.drawOval(10,10,30,30);
        
        G.setPaint(grass);
        
        G.fill(wholeCanvas);
        
        G.setPaint(Color.black);
        for(int i=0;i<ballList.size();i++) {
            ((BallRenderer) ballList.get(i)).drawBall(G);
        }
        g.drawImage(newImg,0,0,this);
    }
    
    /**
     * Add a ball to the canvas for rendering.
     *
     * @param rend  A renderer that should be added.
     */
    public void addBallRenderer(BallRenderer rend) {
        requestFocus();
        selectBall(rend);
        ballList.add(rend);
    }

    /**
     * Remove ball from the canvas.
     *
     * @param rend  A renderer that should be added.
     */
    public void removeBallRenderer(BallRenderer rend) {
        rend.stop();
        ballList.remove(rend);
    }
    
    // inherits doc from World
    public synchronized void addBall(Ball aNewBall) {
        aNewBall.setWorld(this);
        addBallRenderer(new BallRenderer(aNewBall, this));
    }
    
    // inherits doc from World
    public Ball getClosestBall(Ball myBall) {
        Point loc = new Point();
        loc.setLocation(myBall.getX(),myBall.getY());
        Ball b = nearestBall(loc,ballList).getBall();
        return (b == myBall) ? null : b;
    }
    
    // inherits doc from World
    public double getMaxX() {
        return getWidth()/2;
    }
    
    // inherits doc from World
    public double getMaxY() {
        return getHeight()/2;
    }

    /**
     * Get the maximum X coordinate in GUI numbers.
     * 
     * @return the width of the component
     */
    public double getGUIMaxX() {
        return getWidth();
    }
    
    /**
     * Get the maximum Y coordinate in GUI numbers.
     *
     * @return the height of the component
     */
    public double getGUIMaxY() {
        return getHeight();
    }
    
    // inherits doc from World
    public double getMinX() {
        return -getWidth()/2;
    }
    
    // inherits doc from World
    public double getMinY() {
        return -getHeight()/2;
    }
    
    // inherits doc from World
    public synchronized void removeBall(Ball toBeRemoved) {
        for (int i=0; i<ballList.size();i++) {
            if (((BallRenderer)ballList.get(i)).getBall().equals(toBeRemoved)) {
                removeBallRenderer((BallRenderer)ballList.get(i));
            }
        }
    }
    
    // inherits doc from Runnable
    public void run() {
        running = true;
        while (running) {
            repaint();
            try {
                Thread.sleep(200);
            } catch (InterruptedException ie) {
                System.out.println("Ball World thread interrupted");
            }
        }
    }
    
    /**
     * Stop automatic repainting.
     */
    public void stop() {
        running = false;
    }
    
    /**
     * A Handler for mouse events. The two events handled are 
     * mouse pressed and mouse released. <p>Mouse pressed checks to see
     * if it occured within the radius of any balls, and makes a list of
     * balls that contain the point. The closest ball to the point 
     * is then selected from the list of balls containing the point.
     * If no ball contains the point the selected ball is deselected.
     * <p>Mouse released events pass their coordinates to the userClicked
     * method of the selected ball. Thus if the userClicked method sets
     * the location of the ball to the coordinates recieved, the ball can
     * be moved by clicking on the ball and then dragging the mouse to a
     * new location. When the button is released, the ball will teleport
     * to the new location.
     *
     * @param evt the <code>MouseEvent</code> to process
     */
    
    private void processClick(MouseEvent evt) {
        if(evt.getID() == MouseEvent.MOUSE_PRESSED) {
            Vector ballsClicked = new Vector();
            Point clickInGUI = new Point();
            clickInGUI.setLocation(evt.getX(),evt.getY());
            Point click = BallRenderer.transformGUI2World(clickInGUI, this);
            
            for (int i=0;i<ballList.size();i++) {
                BallRenderer thisBall = (BallRenderer)ballList.get(i);
                if (thisBall.containsPoint(click)) {
                    ballsClicked.add(thisBall);
                }
            }
            
            BallRenderer select = nearestBall(click, ballsClicked);
            if (select != null) {
                selectBall(select);
            } else {
                if (selectedBall != null) {
                    selectedBall.setSelected(false);
                    selectedBall = (BallRenderer) null;
                }
            }

        }
        if(evt.getID() == MouseEvent.MOUSE_RELEASED) {
            Point clickInGUI = new Point();
            clickInGUI.setLocation(evt.getX(),evt.getY());
            Point click = BallRenderer.transformGUI2World(clickInGUI,this);
            if (selectedBall != null) {
                selectedBall.getBall().userClicked(click.getX(),click.getY());
            }
        }
        this.requestFocus();
    }
    
    private void processKey(KeyEvent evt) {
        if (selectedBall != null) {
            selectedBall.getBall().userTyped(evt.getKeyChar());
        }
    }
    
    private synchronized BallRenderer nearestBall(Point p, Vector ballRends) {
        
        if (ballRends.size() == 0) {
            return (BallRenderer) null;
        }
        
        BallRenderer closest = (BallRenderer) ballRends.get(0);
        for (Iterator iter = ballRends.iterator(); iter.hasNext();) {
            BallRenderer thisBall = (BallRenderer)iter.next();
            if (closest.distanceTo(p) > thisBall.distanceTo(p)) {
                closest = thisBall;
            }
        }
        
        return closest;
    }
    
    private void selectBall(BallRenderer rend) {
        if (this.selectedBall != null) {
            this.selectedBall.setSelected(false);
        }
        this.selectedBall = rend;
        rend.setSelected(true);
    }
}
